/*******************************************************************************
 * 
 * jKiwi -- A virtual makeover and hairstyler application
 * Copyright (c) 2007-2008 Dan Mihai Ile, Maria Joao Roque Barbado Leal
 * 
 * This file is part of jKiwi.
 * 
 * jKiwi is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 * 
 * jKiwi is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 *******************************************************************************/

package core;

import java.awt.image.BufferedImage;
import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLClassLoader;
import java.net.URLDecoder;
import java.util.ArrayList;
import java.util.Locale;
import java.util.ResourceBundle;

import javax.imageio.ImageIO;

import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.graphics.Font;
import org.eclipse.swt.graphics.ImageData;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.SWT;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.widgets.ProgressBar;
import org.xml.sax.SAXException;
import org.xml.sax.XMLReader;
import org.xml.sax.helpers.XMLReaderFactory;


import blending.BlendingComposite;

import tabs.ColorChooserButton;
import tabs.hair.HairButton;
import tabs.hair.HairColorData;
import tabs.hair.HairColoredSample;
import tabs.hair.HairColorsXMLHandler;
import tabs.hair.HairStyleData;
import tabs.hair.HairStyleXMLHandler;
import utils.Base64;
import utils.ImageUtils;

/**
 * Instances of this class loads all the necessary files and set's
 * the needed variables that are going to be used by the application
 */
public class Loader {
	
	private static final String APPLICATION_NAME = "jKiwi";
	private static final String APPLICATION_VERSION = "0.9.5";

	/**
	 * relative paths to resources used by the program
	 */
	private static final String LOCALE_FILES = "locale/locale";
	private static final String RESOURCES_PATH = "resources";
	private static final String HAIRSTYLES_PATH = RESOURCES_PATH + File.separator + "hairstyles/";
	private static final String HAIRCOLORS_FILE = RESOURCES_PATH + File.separator + "haircolors/colors.xml";
	private static final String PIXMAPS_PATH = RESOURCES_PATH + File.separator + "pixmaps/";

	/**
	 * Components used by loader window
	 */
	private Shell sShell = null;
	private Composite composite = null;
	private Label labelLoading = null;
	private Label labelLoadingInfo = null;
	private ProgressBar progressBar = null;

	/**
	 * array list that contains all Blending Composites available for the application
	 */
	private static ArrayList<BlendingComposite> availableBlendingComposites = null;

	/**
	 * store the user's locale
	 */
	private static Locale currentLocale = null;

	/**
	 * All available translated strings can be retrieved using this ResourceBundle
	 */
	private static ResourceBundle messages;

	/**
	 * hair thumbnail's loaded from XML files, to be used by {@link HairButton}
	 */
	public static final ArrayList<HairStyleData> hairThumbnail = new ArrayList<HairStyleData>();

	/**
	 * hair color images loaded from XML file, to be used by {@link ColorChooserButton}
	 */
	public static final ArrayList<HairColoredSample> hairColorImages = new ArrayList<HairColoredSample>();

	/**
	 * The application's main window
	 */
	private MainWindow mainWindow = null;

	/**
	 * @param args the application's arguments
	 */
	public static void main(String[] args) {
		
		System.setProperty("program.name", APPLICATION_NAME);
		System.setProperty("program.version", APPLICATION_VERSION);

		//TODO detect only one instance of running program when using file->open from operating system

		boolean showHelp = false;
		boolean showVersion = false;
		//check for command arguments
		for (int i = 0; i < args.length; i++) {
			if (args[i].equals("-?") || args[i].equals("--help")) {
				showHelp = true;
			} else
			if (args[i].equals("-v") || args[i].equals("--version")) {
				showVersion = true;
			} else
			//the locale to be used by the program
			if (args[i].equals("--locale") && (args.length > i+1)) {
				String locale = args[i+1];
				if (locale.contains("_")) {
					currentLocale = new Locale(locale.split("_")[0],locale.split("_")[1]);
				} else {
					currentLocale = new Locale(locale);
				}
				i++;
			} else
			if (args[i].equals("--import-photos") && (args.length > i+1)) {
				File f = new File(args[i+1]);
				if (f.exists() && f.isDirectory()) {
					System.setProperty("import-photos", f.getAbsolutePath());
				}
				i++;
			}
		}

		Display display = Display.getDefault();
		Loader thisClass = new Loader();

		//set program's source path (should be renamed to resources path!?)
		if (System.getProperty("program.sourcepath") == null) {
			try {
				//get the jar's path
				String sourcePath = thisClass.getClass().getProtectionDomain().getCodeSource().getLocation().getFile();
				sourcePath = URLDecoder.decode(sourcePath, "utf8");
				if (!new File(sourcePath).isDirectory()) {
					sourcePath = new File(sourcePath).getParent() + File.separator;
				}
				
				//get out of the bin folder
				sourcePath = new File(sourcePath).getParent() + File.separator;
				
				//set the source path property
				System.setProperty("program.sourcepath", sourcePath);
				
			} catch (Exception e) {
				//TODO transfer to locale?! how..
				System.err.println("Unable to get the program's source path");
				return ;
			}
		} else {
			//check if the given path as java argument is a directory
			if (!new File(System.getProperty("program.sourcepath")).isDirectory()) {
				System.err.println("Unable to get the program's source path");
				return ;
			}
			//check if the given path as java argument ends with the File.separator
			if (!System.getProperty("program.sourcepath").endsWith(File.separator)) {
				//fix the source path property
				System.setProperty("program.sourcepath", System.getProperty("program.sourcepath") + File.separator);
			}
		}
		
		System.setProperty("program.resourcespath", System.getProperty("program.sourcepath") + RESOURCES_PATH);
		System.setProperty("program.pixmapspath", System.getProperty("program.sourcepath") + PIXMAPS_PATH);
		System.setProperty("program.icon", System.getProperty("program.pixmapspath") + "program_icon.png");

		//set the current used locale
		if (currentLocale == null) { //the user didn't specified any -locale argument
			currentLocale = new Locale(System.getProperty("user.language"),System.getProperty("user.country"));
		}

		//create a class loader with the program's source path to load the locale 
		try {
			URLClassLoader localesLoader = URLClassLoader.newInstance(new URL [] {new URL("file:" + System.getProperty("program.sourcepath"))});
			messages = ResourceBundle.getBundle(LOCALE_FILES,currentLocale, localesLoader);
		} catch (MalformedURLException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

		if (!showHelp && !showVersion) {
			initializeBlendingModes();
			thisClass.createSShell();

/*
 * Java 6 Splash screen is not being used, a better implementation is needed.
 * (when in compiz the loading screen gets two times the open effect.)
			try { //test for SplashScreen and close it
				Class.forName("java.awt.SplashScreen");
				if (java.awt.SplashScreen.getSplashScreen() != null) {
					java.awt.SplashScreen.getSplashScreen().close();
				}
			} catch (ClassNotFoundException e) {
				//SplashScreen class not available, no problem
			}
*/
			while (!thisClass.sShell.isDisposed()) {
				if (!display.readAndDispatch())
					display.sleep();
			}
			display.dispose();
		} else {
			if (showHelp) {
				printHelp();
			} else
			if (showVersion) {
				System.out.println(APPLICATION_VERSION);
			}
		}
	}

	private static void printHelp() {
		System.out.println(messages.getString("TERMINAL_HELP_USAGE"));
		System.out.println("  " + messages.getString("TERMINAL_HELP_JKIWI"));
		System.out.println("\n" + messages.getString("TERMINAL_HELP_OPTIONS"));
		System.out.println("  -?, --help" + messages.getString("TERMINAL_HELP_OPTION1"));
		System.out.println("\n" + messages.getString("TERMINAL_HELP_APPLICATION_OPTIONS"));
		System.out.println("  -locale LOCALE" + messages.getString("TERMINAL_HELP_APPLICATION_OPTION1"));
		System.out.println("  --import-photos PATH" + messages.getString("TERMINAL_HELP_APPLICATION_OPTION2"));
		System.out.println("  -v, --version" + messages.getString("TERMINAL_VERSION_OPTION1"));
		System.out.println();
	}

	/**
	 * This method initializes composite	
	 * @param mainComposite 
	 *
	 */
	private void createComposite(Composite mainComposite) {

		//grid data for the composite
		GridData gridDataComposite = new GridData();
		gridDataComposite.horizontalSpan = 2;
		gridDataComposite.verticalAlignment = org.eclipse.swt.layout.GridData.END;
		gridDataComposite.grabExcessHorizontalSpace = true;
		gridDataComposite.horizontalAlignment = org.eclipse.swt.layout.GridData.FILL;
		gridDataComposite.grabExcessVerticalSpace = true;

		//initialize the composite
		composite = new Composite(mainComposite, SWT.NONE);
		composite.setBackground(Display.getCurrent().getSystemColor(SWT.COLOR_WHITE));
		composite.setLayoutData(gridDataComposite);
		composite.setBackgroundImage(new Image(Display.getCurrent(), System.getProperty("program.pixmapspath") + "loading.png"));

		//adjust the composite size to it's background image size
		gridDataComposite.widthHint = composite.getBackgroundImage().getBounds().width;
		gridDataComposite.heightHint = composite.getBackgroundImage().getBounds().height;

	}

	/**
	 * All blending composites that are going to be used by the application are specified here
	 */
	private static void initializeBlendingModes() {
		availableBlendingComposites = new ArrayList<BlendingComposite>();
		availableBlendingComposites.add(new BlendingComposite(BlendingComposite.GRAIN_MERGE));
		availableBlendingComposites.add(new BlendingComposite(BlendingComposite.MULTIPLY));
		availableBlendingComposites.add(new BlendingComposite(BlendingComposite.COLOR_OVERLAY));
	}

	/**
	 * Return the composite for the desired blending mode.
	 * In case that no composite is found a <code>null</code>
	 * value is returned.
	 * @param blendingMode the desired blending mode
	 * @return the composite for the blending mode or a <code>null</code> value
	 */
	public static java.awt.Composite getCompositeFor(int blendingMode) {
		for (BlendingComposite bc : availableBlendingComposites) {
			if (bc.getBlendingMode() == blendingMode) {
				return (java.awt.Composite) bc;
			}
		}
		return null;
	}

	/**
	 * Method that returns the messages that are to be used by the entire GUI
	 * @return the messages
	 */
	public static ResourceBundle getMessages() {
		return messages;
	}

	/**
	 * This method initializes sShell
	 * @param args program arguments
	 */
	private void createSShell() {

		//grid layout for the white composite
		GridLayout gridLayoutWhiteComposite = new GridLayout();
		gridLayoutWhiteComposite.marginWidth = 1;
		gridLayoutWhiteComposite.marginHeight = 1;
		gridLayoutWhiteComposite.horizontalSpacing = 0;
		gridLayoutWhiteComposite.verticalSpacing = 0;

		//grid layout for the shell
		GridLayout gridLayoutSShell = new GridLayout();
		gridLayoutSShell.marginWidth = 1;
		gridLayoutSShell.marginHeight = 1;
		gridLayoutSShell.horizontalSpacing = 0;
		gridLayoutSShell.verticalSpacing = 0;

		//grid layout for the main composite
		GridLayout gridLayoutMainComposite = new GridLayout();
		gridLayoutMainComposite.numColumns = 2;
		gridLayoutMainComposite.marginWidth = 1;
		gridLayoutMainComposite.marginHeight = 1;
		gridLayoutMainComposite.marginWidth = 1;
		gridLayoutMainComposite.marginHeight = 1;
		gridLayoutMainComposite.horizontalSpacing = 2;
		gridLayoutMainComposite.verticalSpacing = 2;

		//grid data for the progress bar
		GridData gridDataProgressBar = new GridData();
		gridDataProgressBar.grabExcessHorizontalSpace = true;
		gridDataProgressBar.horizontalSpan = 2;
		gridDataProgressBar.horizontalAlignment = org.eclipse.swt.layout.GridData.FILL;
		gridDataProgressBar.heightHint = 15;

		//grid data for the label
		GridData gridDataLabel = new GridData();
		gridDataLabel.grabExcessHorizontalSpace = true;
		gridDataLabel.horizontalAlignment = org.eclipse.swt.layout.GridData.FILL;

		//initialize the shell
		Shell starter = new Shell(); //avoid the loader shell to appear in window list, by passing this starter shell as argument  
		sShell = new Shell(starter, SWT.NO_TRIM);
		sShell.setLayout(gridLayoutSShell);
		sShell.setBackground(Display.getCurrent().getSystemColor(SWT.COLOR_GRAY));

		//initialize the white composite
		Composite whiteComposite = new Composite(sShell, SWT.NONE);
		whiteComposite.setBackground(Display.getCurrent().getSystemColor(SWT.COLOR_WHITE));
		whiteComposite.setLayout(gridLayoutWhiteComposite);
		whiteComposite.setLayoutData(new GridData());

		//initialize the main composite
		Composite mainComposite = new Composite(whiteComposite, SWT.NONE);
		mainComposite.setLayout(gridLayoutMainComposite);
		mainComposite.setLayoutData(new GridData());
		mainComposite.setBackground(Display.getCurrent().getSystemColor(SWT.COLOR_BLACK));
		createComposite(mainComposite);

		//initialize the loading label
		labelLoading = new Label(mainComposite, SWT.NONE);
		labelLoading.setText(messages.getString("LOADER-LOADING"));
		labelLoading.setForeground(Display.getCurrent().getSystemColor(SWT.COLOR_WHITE));
		labelLoading.setBackground(Display.getCurrent().getSystemColor(SWT.COLOR_BLACK));
		labelLoading.setFont(new Font(Display.getDefault(), "Sans", 9, SWT.NORMAL));

		//initialize the label with the loading info
		labelLoadingInfo = new Label(mainComposite, SWT.NONE);
		labelLoadingInfo.setBackground(Display.getCurrent().getSystemColor(SWT.COLOR_BLACK));
		labelLoadingInfo.setForeground(Display.getCurrent().getSystemColor(SWT.COLOR_WHITE));
		labelLoadingInfo.setLayoutData(gridDataLabel);
		labelLoadingInfo.setFont(new Font(Display.getDefault(), "Sans", 9, SWT.NORMAL));

		//initialize the progress bar
		progressBar = new ProgressBar(mainComposite, SWT.SMOOTH);
		progressBar.setLayoutData(gridDataProgressBar);

		sShell.pack(); //pack is done before setting the shell location

		int width = Display.getCurrent().getBounds().width;
		int height = Display.getCurrent().getBounds().height;
/*
 * No java6 splash screen for now
		//open the shell in the center of the screen (the loading image is used to align to center because of -splash)
		//-4 and -3 values are because of -splash option from java6
		sShell.setLocation(
				(width / 2) - (composite.getBackgroundImage().getBounds().width / 2) - 4,
				(height / 2) - (composite.getBackgroundImage().getBounds().height / 2) - 3
		);
*/
		sShell.setLocation(
				(width / 2) - (composite.getBackgroundImage().getBounds().width / 2),
				(height / 2) - (composite.getBackgroundImage().getBounds().height / 2)
		);
		sShell.open();

		//create and start the worker thread
		Worker worker = new Worker(Display.getCurrent());
		worker.start();
	}

	/**
	 * Worker thread meant to do all the resource loading
	 */
	private class Worker extends Thread {

		/**
		 * XML reader used to read XML hair styles and XML hair colors files
		 */
		private XMLReader parser = null;

		/**
		 * counter that keeps track of the current number of XML hair style files loaded
		 */
		private int xmlFilesLoaded = 0;

		/**
		 * the display in use by the loader shell
		 */
		private Display display = null;

		/**
		 * Create a new instance of this Worker
		 * @param current the used display
		 */
		public Worker(Display current) {
			display = current;
		}

		@Override
		public void run() {

			//set the cursor to wait
			display.asyncExec(new Runnable() {
				public void run() {
					sShell.setCursor(display.getSystemCursor(SWT.CURSOR_WAIT));
				}
			});

			//initialize the XML parser with an available reader
			try {
				parser = XMLReaderFactory.createXMLReader();
			} catch (SAXException e1) {
				// TODO Auto-generated catch block
				e1.printStackTrace();
			}

			loadHairColors();
			loadHairStyles();
			loadMainWindow();

			if (!display.isDisposed()) {
				display.asyncExec(new Runnable() {
					//restore the default cursor
					//TODO is this really needed?
					public void run() {
						sShell.setCursor(display.getSystemCursor(SWT.DEFAULT));
					}
				});
			}
		}

		/**
		 * open the main window and close the Loader window
		 */
		private void loadMainWindow() {
			if (display.isDisposed()) return ;

			display.asyncExec(new Runnable() {
				public void run() {
					labelLoadingInfo.setText(messages.getString("MAIN_WINDOW_TITLE"));
					mainWindow =  new MainWindow();

					mainWindow.getSShell().open();
					Shell currentShell = sShell;
					sShell = mainWindow.getSShell();
					currentShell.close();
				}
			});
		}

		/**
		 * Load the hair colors from the XML hair colors file
		 */
		private void loadHairColors() {

			//update the loading info label
			display.asyncExec(new Runnable() {
				public void run() {
					if (!progressBar.isDisposed()) {
						labelLoadingInfo.setText(messages.getString("LOADER-HAIR_COLORS_XML_FILE"));
					}
				}
			});

			//create a handler for parsing the XML file
			HairColorsXMLHandler hairColorsHandler = new HairColorsXMLHandler();
			parser.setContentHandler(hairColorsHandler);

			//load the hair sample image
			BufferedImage hairColorSampleImage = null;
			try {
				hairColorSampleImage = ImageIO.read(new File(System.getProperty("program.pixmapspath") + "hair_color.png"));
			} catch (IOException e1) {
				// TODO Auto-generated catch block
				e1.printStackTrace();
			}

			//parse the XML file
			try {
				parser.parse(new File(System.getProperty("program.sourcepath") + HAIRCOLORS_FILE).toURI().toString());
				for (HairColorData hairColor : hairColorsHandler.getHairColorsList()) {
					Image im = new Image(Display.getCurrent(),ImageUtils.applyBlendingSWT(hairColorSampleImage, getCompositeFor(hairColor.getBlendingMode()), hairColor.getRed(), hairColor.getGreen(), hairColor.getBlue()));
					hairColorImages.add(new HairColoredSample(hairColor,im));
				}
			} catch (IOException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			} catch (SAXException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}

		/**
		 * Count the number of XML files and call for the loading of
		 * all the hair style XML files found in the HAIRSTYLES_PATH
		 */
		private void loadHairStyles() {

			//create a handler for parsing the XML file
			HairStyleXMLHandler hairStyleHandler = new HairStyleXMLHandler(HairStyleXMLHandler.FOR_THUMBNAIL_READING);
			parser.setContentHandler(hairStyleHandler);

			//initialize the parsing directory and count the number of files to be processed
			File hairstylesDirectory = new File(System.getProperty("program.sourcepath") + HAIRSTYLES_PATH);
			final double numFiles = listAllDirectoryFiles(hairstylesDirectory);// hairstylesDirectory.listFiles(new IsFile()).length;

			//update the loading info label
			display.asyncExec(new Runnable() {
				public void run() {
					if (!progressBar.isDisposed()) {
						labelLoadingInfo.setText(messages.getString("LOADER-HAIR_STYLE_XML_FILE"));
					}
				}
			});
			//call for reading of all XML hair files
			readXMLHairStyles(hairStyleHandler, hairstylesDirectory, numFiles);
		}

		/**
		 * Load all XML hair style files found in hairstylesDirectory,
		 * using the given hairStyleHandler while updating the progress bar
		 * @param hairStyleHandler the handler to be used for the parsing
		 * @param hairstylesDirectory the directory in which the XML files are located
		 * @param numFiles the total number of XML files to be processed
		 */
		private void readXMLHairStyles(HairStyleXMLHandler hairStyleHandler, File hairstylesDirectory, final double numFiles) {

			//update the progress bar
			Runnable updateWindow = new Runnable() {
				public void run() {
					if (!progressBar.isDisposed()) {
						progressBar.setSelection((int)((xmlFilesLoaded / numFiles) * 100));
					}
				}
			};

			//parse every file in the directory or recursively jump to the next directory
			for (final File f : hairstylesDirectory.listFiles()) {
				if (f.isFile()) {
					try {
						hairStyleHandler.initReadingVars(); //reset vars before reading
						parser.parse(f.toURI().toString());
						Image thumbnail = new Image(Display.getCurrent(), new ImageData(new ByteArrayInputStream (Base64.decode(hairStyleHandler.getStrThumbnail()))));
						HairStyleData hs = new HairStyleData(hairStyleHandler.getHairStyle_title(), hairStyleHandler.getHairOwner_name(), hairStyleHandler.getHairOwner_gender(), hairStyleHandler.getHairStyle_type(), thumbnail, f);
						hairThumbnail.add(hs);
						xmlFilesLoaded ++;
					} catch (Exception e) {
						System.out.println(messages.getString("LOADER-ERROR_LOADING_XML_FILE") + ": " + f);
					}
					if (display.isDisposed()) {
						return;
					}
					display.asyncExec(updateWindow);
				} else if (f.isDirectory() && !f.getName().equals("CVS")) {
					readXMLHairStyles(hairStyleHandler, f, numFiles);
				}
			}
		}
	}

	/**
	 * Process the given directory and count all the files in it
	 * @param hairstylesDirectory the desired directory
	 * @return the number of files found in it
	 */
	private int listAllDirectoryFiles(File hairstylesDirectory) {
		int res = 0;
		for (File f : hairstylesDirectory.listFiles()) {
			if (f.isDirectory() && !f.getName().equals("CVS")) {
				res += listAllDirectoryFiles(f);
			} else {
				if (f.isFile()) {
					res++;
				}
			}
		}
		return res;
	}
	
	/**
	 * used to print to terminal the difference between the end and start time 
	 */
	public static String printTime(long start, long end) {

		String finalTimeRepresentation = "";
		long totalMilliseconds = end - start;
		long milliseconds = totalMilliseconds % 1000;
		long timeInSeconds = totalMilliseconds / 1000;
		long hours, minutes, seconds;
		hours = timeInSeconds / 3600;
		timeInSeconds = timeInSeconds - (hours * 3600);
		minutes = timeInSeconds / 60;
		timeInSeconds = timeInSeconds - (minutes * 60);
		seconds = timeInSeconds;

		boolean gotTime = false;
		if (hours != 0) {
			finalTimeRepresentation += (hours > 1)?(hours + " hours "):(hours + " hour ");
			gotTime = true;
		}
		if (minutes != 0) {
			finalTimeRepresentation += (minutes > 1)?(minutes + " minutes "):(minutes + " minute ");
			gotTime = true;
		}
		if (seconds != 0) {
			finalTimeRepresentation += (seconds > 1)?(seconds + " seconds "):(seconds + " second ");
			gotTime = true;
		}
		if (milliseconds != 0 || !gotTime) {
			if (gotTime) {
				finalTimeRepresentation += "and ";
			}
			finalTimeRepresentation += (milliseconds == 1)?(milliseconds + " millisecond"):(milliseconds + " milliseconds");
		}

		if (finalTimeRepresentation.endsWith(" ")) {
			finalTimeRepresentation = finalTimeRepresentation.substring(0, finalTimeRepresentation.length() - 1);
		}

		return finalTimeRepresentation;
	}
}