/*******************************************************************************
 * 
 * jKiwi -- A virtual makeover and hairstyler application
 * Copyright (c) 2007-2008 Dan Mihai Ile, Maria Joao Roque Barbado Leal
 * 
 * This file is part of jKiwi.
 * 
 * jKiwi is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 * 
 * jKiwi is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 *******************************************************************************/

package dialogs;

import java.awt.image.BufferedImage;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.Iterator;
import java.util.ResourceBundle;

import javax.imageio.ImageIO;
import javax.imageio.ImageReader;
import javax.imageio.ImageWriter;
import javax.imageio.event.IIOReadProgressListener;
import javax.imageio.event.IIOWriteProgressListener;
import javax.imageio.stream.FileImageInputStream;
import javax.imageio.stream.FileImageOutputStream;
import javax.swing.SwingUtilities;


import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.widgets.MessageBox;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.SWT;
import org.eclipse.swt.widgets.ProgressBar;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.widgets.Display;

import core.Loader;

import utils.StockIcon;

/**
 * Instances of this class creates a new dialog
 * showing the progress of an IO operation
 */
public class FileIOProgress {

	/**
	 * The shell used by this dialog
	 */
	private Shell sShell = null;

	/**
	 * messages that are to be used by the entire GUI
	 */
	private ResourceBundle messages;

	/**
	 * The label to be used for showing the dialog message
	 */
	private Label label = null;

	/**
	 * The progress bar that goes from 0 to 100
	 */
	private ProgressBar progressBar = null;

	/**
	 * The cancel button
	 */
	private Button button = null;

	/**
	 * Flag used to signal that cancel was pressed
	 * or that the window was closed before the reading ends
	 */
	private boolean cancelOperation = false;

	/**
	 * The image that has just been read by openImage method
	 */
	private BufferedImage readedImage = null;

	/**
	 * The reader that is used to read images
	 */
	private ImageReader imageReader = null;

	/**
	 * Flag used to signal that the image was saved successfully to the specified file
	 */
	private boolean imageSavedSuccessfully = false;

	/**
	 * Creates a new instance of FileIOProgress dialog
	 * given it's parent Shell
	 * @param sShell the Shell to be used as modal
	 */
	public FileIOProgress(Shell sShell) {
		messages = Loader.getMessages();
		createSShell(sShell);
	}

	/**
	 * This method initializes sShell
	 */
	private void createSShell(Shell shell) {

		//label grid data
		GridData gridDataLabel = new GridData();
		gridDataLabel.horizontalSpan = 2;
		gridDataLabel.horizontalAlignment = org.eclipse.swt.layout.GridData.FILL;
		gridDataLabel.grabExcessHorizontalSpace = true;

		//button grid data
		GridData gridDataButton = new GridData();
		gridDataButton.grabExcessHorizontalSpace = true;
		gridDataButton.horizontalAlignment = org.eclipse.swt.layout.GridData.END;
		gridDataButton.verticalAlignment = org.eclipse.swt.layout.GridData.END;
		gridDataButton.grabExcessVerticalSpace = true;
		gridDataButton.horizontalSpan = 2;

		//progress bar grid data
		GridData gridDataProgressBar = new GridData();
		gridDataProgressBar.horizontalSpan = 2;
		gridDataProgressBar.horizontalAlignment = org.eclipse.swt.layout.GridData.FILL;
		gridDataProgressBar.grabExcessHorizontalSpace = true;

		//shell grid layout
		GridLayout gridLayoutSShell = new GridLayout();
		gridLayoutSShell.numColumns = 2;
		gridLayoutSShell.marginWidth = 10;
		gridLayoutSShell.marginHeight = 10;

		//create and initialize the shell
		sShell = new Shell(shell, SWT.APPLICATION_MODAL | SWT.DIALOG_TRIM);
		sShell.setText(messages.getString("FILE-IO-PROGRESS_WINDOW_TITLE"));
		sShell.setLayout(gridLayoutSShell);
		sShell.setSize(new Point(513, 155));

		//create the label, progress bar and button in the shell
		label = new Label(sShell, SWT.NONE);
		label.setLayoutData(gridDataLabel);
		progressBar = new ProgressBar(sShell, SWT.NONE);
		progressBar.setLayoutData(gridDataProgressBar);
		button = new Button(sShell, SWT.NONE);
		button.setText(messages.getString("FILE-IO-PROGRESS_CANCEL_BUTTON"));
		button.setLayoutData(gridDataButton);
		button.setImage(new Image(Display.getCurrent(), StockIcon.getIcon(StockIcon.ICON_CANCEL, StockIcon.SIZE_LARGE_TOOLBAR, true)));

		//button and shell listeners
		button.addSelectionListener(new org.eclipse.swt.events.SelectionAdapter() {
			public void widgetSelected(org.eclipse.swt.events.SelectionEvent e) {
				cancelOperation = true;
			}
		});
		sShell.addShellListener(new org.eclipse.swt.events.ShellAdapter() {
			public void shellClosed(org.eclipse.swt.events.ShellEvent e) {
				cancelOperation = true;
			}
		});

		//set the shell's default button to cancel button
		sShell.setDefaultButton(button);
	}

	/**
	 * Save a given image to png format in the given file
	 * @param image the image to save
	 * @param file the file used to save the image
	 * @return true if save was successful
	 */
	public boolean saveImage(final BufferedImage image, final String file) {

		imageSavedSuccessfully = false;
		label.setText(messages.getString("FILE-IO-PROGRESS_LABEL_TEXT_SAVE") + " '" + file + "'..."); //set the label's text
		sShell.pack(); //pack the window to ensure no text is cut from the window
		sShell.open(); //open the window

		//create a progress listener so that we get info about the image writing progress
		final IIOWriteProgressListener progressListener = new IIOWriteProgressListener() {
			//unimplemented methods
			public void imageStarted(ImageWriter source, int imageIndex) {}
			public void thumbnailComplete(ImageWriter source) {}
			public void thumbnailProgress(ImageWriter source, float percentageDone) {}
			public void thumbnailStarted(ImageWriter source, int imageIndex, int thumbnailIndex) {}
			public void writeAborted(ImageWriter source) {}

			// Set image progress to 100% upon completion.

			public void imageComplete(ImageWriter source) {
				imageSavedSuccessfully = true;
				imageProgress(source, 100.0F);
			}

			// Update the progress bar and its label each time the writer
			// notifies the IIOWriteProgressListener of a new percentage.
			public void imageProgress(ImageWriter source, float percentageDone) {
				SwingUtilities.invokeLater(guiUpdater(percentageDone, source));
			}
		};

		//create a worker thread to open the image
		new Thread() {
			public void run() {

				Iterator<ImageWriter> writers = ImageIO.getImageWritersByFormatName("png");
				ImageWriter imageWriter = writers.next();
				imageWriter.addIIOWriteProgressListener(progressListener);
				try {
					imageWriter.setOutput(new FileImageOutputStream(new File(file)));
//					ImageWriteParam param = imageWriter.getDefaultWriteParam();
//					param.setCompressionMode(ImageWriteParam.MODE_EXPLICIT);
//					param.setCompressionQuality(1.0f);
					imageWriter.write(image);
				} catch (FileNotFoundException e) {
					sShell.getDisplay().syncExec(new Runnable() {
						public void run() {
							String title = messages.getString("FILE-IO-PROGRESS_MESSAGE-BOX_SAVE_CREATEFILE_TITLE");
							String message = messages.getString("FILE-IO-PROGRESS_MESSAGE-BOX_SAVE_CREATEFILE_MESSAGE") + " \"" + file + "\"";
							showIOReadingError(SWT.ICON_ERROR, title, message);
							sShell.close();
						}
					});
				} catch (IOException e) {
					sShell.getDisplay().syncExec(new Runnable() {
						public void run() {
							String title = messages.getString("FILE-IO-PROGRESS_MESSAGE-BOX_SAVE_IOEXCEPTION_TITLE");
							String message = messages.getString("FILE-IO-PROGRESS_MESSAGE-BOX_SAVE_IOEXCEPTION_MESSAGE") + " \"" + file + "\"";
							showIOReadingError(SWT.ICON_ERROR, title, message);
							sShell.close();
						}
					});
				}
			}
		}.start();

		//while the worker thread is not finished reading the image update GUI and then sleep
		while (!sShell.isDisposed()) {
			if (!sShell.getDisplay().readAndDispatch())
				sShell.getDisplay().sleep();
		}
		//if cancel was pressed then try to delete the image file created
		if (cancelOperation) {
			new File(file).delete();
		}
		return imageSavedSuccessfully;
	}

	/**
	 * Method that opens an image (jpeg or png)
	 * @param file the file that is to be read
	 * @return an BufferedImage with the image just read
	 */
	public BufferedImage openImage(final String file) {
		if (file == null) return null;
		readedImage = null; //initialize the resulting image as null
		label.setText(messages.getString("FILE-IO-PROGRESS_LABEL_TEXT_OPEN") + " '" + file + "'..."); //set the label's text
		sShell.pack(); //pack the window to ensure no text is cut from the window
		sShell.open(); //open the window

		//create a progress listener so that we get info about the image reading progress
		final IIOReadProgressListener progressListener = new IIOReadProgressListener() {
			//unimplemented methods
			public void readAborted(ImageReader source) {}
			public void imageStarted(ImageReader source, int imageIndex) {}
			public void thumbnailStarted(ImageReader source, int imageIndex,int thumbnailIndex) {}
			public void thumbnailProgress(ImageReader source, float percentageDone) {}
			public void thumbnailComplete(ImageReader source) {}
			public void sequenceStarted(ImageReader source, int minIndex) {}
			public void sequenceComplete(ImageReader source) {}

			// Set image progress to 100% upon completion.
			public void imageComplete(ImageReader source) {
				imageProgress(source, 100.0F);
			}

			// Update the progress bar and its label each time the reader
			// notifies the IIOReadProgressListener of a new percentage.
			public void imageProgress(ImageReader source, float percentageDone) {
				SwingUtilities.invokeLater(guiUpdater(percentageDone, source));
			}
		};

		//create a worker thread to open the image
		new Thread() {
			public void run() {
				try { //first try to read the file as an jpeg file
					Iterator<ImageReader> readers = ImageIO.getImageReadersByFormatName("jpeg");
					imageReader = readers.next();
					imageReader.addIIOReadProgressListener(progressListener);
					imageReader.setInput(new FileImageInputStream(new File(file)));
					readedImage = imageReader.read(0);
				} catch (IOException e1) { //image could not be of type jpeg, try png instead
					try {
						Iterator<ImageReader> readers = ImageIO.getImageReadersByFormatName("png");
						imageReader = readers.next();
						imageReader.addIIOReadProgressListener(progressListener);
						imageReader.setInput(new FileImageInputStream(new File(file)));
						readedImage = imageReader.read(0);
					} catch (IOException e) {
						if (!(new File(file).exists())) { //file not found error
							sShell.getDisplay().syncExec(new Runnable() {
								public void run() {
									String title = messages.getString("FILE-IO-PROGRESS_MESSAGE-BOX_NOTFOUND_TITLE");
									String message = messages.getString("FILE-IO-PROGRESS_MESSAGE-BOX_NOTFOUND_MESSAGE1") + 
									" \"" + file + "\"" + ".\n" + messages.getString("FILE-IO-PROGRESS_MESSAGE-BOX_NOTFOUND_MESSAGE2") + ".";
									showIOReadingError(SWT.ICON_ERROR, title, message);
								}
							});
						} else { //other error occurred
							sShell.getDisplay().syncExec(new Runnable() {
								public void run() {
									String title = messages.getString("FILE-IO-PROGRESS_MESSAGE-BOX_ERROR_TITLE");
									String message = messages.getString("FILE-IO-PROGRESS_MESSAGE-BOX_ERROR_MESSAGE1") + ".\n" +
									messages.getString("FILE-IO-PROGRESS_MESSAGE-BOX_ERROR_MESSAGE2") + ".";
									showIOReadingError(SWT.ICON_WARNING, title, message);
								}
							});
						}
						sShell.getDisplay().syncExec(new Runnable() {
							public void run() {
								sShell.close();
							}
						});
					}
				} catch (OutOfMemoryError e) {
					sShell.getDisplay().syncExec(new Runnable() {
						public void run() {
							String title = messages.getString("FILE-IO-PROGRESS_MESSAGE-BOX_OUTOFMEMORY_TITLE");
							String message = messages.getString("FILE-IO-PROGRESS_MESSAGE-BOX_OUTOFMEMORY_MESSAGE") + ".";
							showIOReadingError(SWT.ICON_ERROR, title, message);
							sShell.close();
						}
					});
				}
			}
		}.start();

		//while the worker thread is not finished reading the image update GUI and then sleep
		while (!sShell.isDisposed()) {
			if (!sShell.getDisplay().readAndDispatch())
				sShell.getDisplay().sleep();
		}
		//if cancel was not pressed then make sure that the image is returned instead of null
		if (!cancelOperation) {
			while (readedImage == null)
				try {
					Thread.sleep(10);
				} catch (InterruptedException e) {
					e.printStackTrace();
				}
		}
//		imageReader.dispose(); //TODO with this line the cancel button does not work!!!
		sShell.dispose(); //TODO is it necessary?
//		imageReader = null;
		return readedImage; //return the image read by the worker thread
	}

	/**
	 * Display a message box given its icon,
	 * a message box title an the content message
	 * @param swtIcon the icon used by the message box
	 * @param title title of the message box
	 * @param message the content of the message box
	 */
	private void showIOReadingError(int swtIcon, String title, String message) {
		//notify the user in a new thread with the given message
		MessageBox mb = new MessageBox(sShell, swtIcon);
		mb.setText(title);
		mb.setMessage(message);
		mb.open();
	}

	/**
	 * Returns a new Runnable used to update the dialog used by this class
	 * @param percentageDone the percentage used by the progress bar
	 * @param sourceObject the object that is responsible for this operation (ImageReader or ImageWriter)
	 * @return the new created Runnable
	 */
	private Runnable guiUpdater(final float percentageDone, final Object sourceObject) {
		return new Runnable() {
			public void run() {
				if (sShell.isDisposed() || progressBar.isDisposed()) {
					sourceAbort(sourceObject);
					return;
				}
				sShell.getDisplay().asyncExec(new Runnable() {
					public void run() {
						if (!progressBar.isDisposed()) {
							progressBar.setSelection(Math.round(percentageDone));
						}
						if (cancelOperation) {
							sourceAbort(sourceObject);
							sShell.dispose();
						}
						if (percentageDone == 100.0) {
							sShell.dispose();
						}
					}
				});
			}
		};
	}

	/**
	 * Tells the given object to abort it's operation
	 * 
	 * @see ImageReader
	 * @see ImageWriter
	 * 
	 * @param sourceObject the object that should abort it's operation
	 */
	private void sourceAbort(Object sourceObject) {
		if (sourceObject instanceof ImageReader) {
			ImageReader source = (ImageReader) sourceObject;
			source.abort();
		} else if (sourceObject instanceof ImageWriter) {
			ImageWriter source = (ImageWriter) sourceObject;
			source.abort();
		}
	}
}